//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef EDITOR_H
#define EDITOR_H

#ifdef _MSC_VER
	#pragma once
#endif

#include <vd2/system/refcount.h>
#include <vd2/system/VDString.h>
#include <vd2/plugin/vdvideofilt.h>
#include "effect.h"

class IVDXFilterPreview;
class IVDXFilterPreview2;
class VDStringA;
class IVDShaderFilter;

class IVDShaderFilterSource {
public:
	virtual void ReadSource(VDStringA& dst) = 0;
};

struct VDShaderFilterDefinition {
	VDXFilterDefinition	mFilterDef;
	VDXFilterDefinition	*mpActiveFilterDef;
	VDStringA mName;
	VDStringA mAuthor;
	VDStringA mDescription;
	VDStringW mShaderPath;
};

enum VDFXParamType {
	kVDFXParamType_None,
	kVDFXParamType_Float,
	kVDFXParamType_Float2,
	kVDFXParamType_Float3,
	kVDFXParamType_Float4,
	kVDFXParamType_Count
};

struct VDShaderFilterParameterConfig {
	VDStringA			mName;
	VDFXParamType		mType;
	VDShaderConstant	mValue;
};

struct VDShaderFilterConfig {
	enum {
		kModeDefault,
		kModeInterp,
		kModeScalarJIT,
		kModeSSE2JIT,
		kModeCount
	};

	VDShaderFilterConfig()
		: mExecutionMode(kModeDefault)
		, mbOptimizationEnabled(true)
		, mbKeepSize(true)
		, mNewWidth(0)
		, mNewHeight(0)
	{
	}

	int mExecutionMode;
	bool mbOptimizationEnabled;
	bool	mbKeepSize;
	int		mNewWidth;
	int		mNewHeight;

	VDStringA mFilename;

	vdrefptr<VDShaderEffect> mpEffect;

	typedef std::vector<VDShaderFilterParameterConfig> ParamValues;
	ParamValues mParamValues;

	IVDShaderFilterSource *mpSource;
};

void VDRegisterShaderEditor();
void VDUnregisterShaderEditor();
bool VDOpenShaderEditor(VDXHWND hwnd, VDShaderFilterConfig& config, IVDXFilterPreview *ifp, IVDXFilterPreview2 *ifp2, IVDShaderFilter *filter);

#endif
