//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include <windows.h>
#include "jit.h"

#if 0
VDJITAllocator::VDJITAllocator()
	: mNextChunk(mFreeChunks.end())
{
	SYSTEM_INFO si;
	GetSystemInfo(&si);

	mAllocationGranularity = si.dwAllocationGranularity;
}

VDJITAllocator::~VDJITAllocator() {
	for(Allocations::iterator it(mAllocations.begin()), itEnd(mAllocations.end()); it!=itEnd; ++it) {
		VirtualFree(it->first, 0, MEM_RELEASE);
	}
}

void *VDJITAllocator::Allocate(size_t len) {
	len = (len + 15) & ~(size_t)15;

	FreeChunks::iterator itMark(mNextChunk), itEnd(mFreeChunks.end()), it(itMark);

	for(;;) {
		for(; it!=itEnd; ++it) {
			if (it->second >= len) {
				it->second -= len;

				void *p = (char *)it->first + it->second;

				if (!it->second) {
					if (mNextChunk == it)
						++mNextChunk;

					mFreeChunks.erase(it);
				}

				return p;
			}
		}

		if (itEnd == itMark)
			break;

		it = mFreeChunks.begin();
		itEnd = itMark;
	}

	size_t alloclen = (len + mAllocationGranularity - 1) & ~(mAllocationGranularity - 1);

	void *p = VirtualAlloc(NULL, alloclen, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
	if (p) {
		try {
			Allocations::iterator itA(mAllocations.insert(Allocations::value_type(p, alloclen)).first);

			try {
				if (alloclen < len)
					mFreeChunks.insert(FreeChunks::value_type((char *)p + len, alloclen - len));

			} catch(...) {
				mAllocations.erase(itA);
			}
		} catch(...) {
			VirtualFree(p, 0, MEM_RELEASE);
		}
	}

	return p;
}

void VDJITAllocator::Free(void *p, size_t len) {
	FreeChunks::iterator cur(mFreeChunks.lower_bound(p));
	if (cur != mFreeChunks.end() && (char *)p + len == cur->first) {
		p = cur->first;
		len += cur->second;
		if (mNextChunk == cur)
			++mNextChunk;
		cur = mFreeChunks.erase(cur);
	}

	if (cur != mFreeChunks.begin()) {
		FreeChunks::iterator prev(cur);

		--prev;
		if ((char *)prev->first + prev->second == p) {
			p = prev->first;
			len += prev->second;
			if (mNextChunk == cur)
				++mNextChunk;
			mFreeChunks.erase(prev);
		}
	}

	uintptr start = (size_t)p;
	uintptr end = start + len;

	if (!((start | end) & (mAllocationGranularity - 1))) {
		Allocations::iterator it(mAllocations.find(p));

		if (it != mAllocations.end()) {
			VirtualFree((void *)start, 0, MEM_RELEASE);
			mAllocations.erase(it);
			return;
		}
	}

	mFreeChunks.insert(FreeChunks::value_type((void *)start, end-start));
}

void VDJITAllocator::EndUpdate(void *p, size_t len) {
	FlushInstructionCache(GetCurrentProcess(), p, len);
}
#endif

namespace x86 {
	BasicBlock::BasicBlock()
		: mpAddress(NULL)
		, mOffset(-1)
		, mpTargetTrue(NULL)
		, mpTargetFalse(NULL)
	{
	}

	BasicBlock::~BasicBlock() {
	}

	void BasicBlock::encode(uint32 opcode, uint32 flags) {
		CreateNewInsn(opcode, flags);
	}

	void BasicBlock::encode_imm(uint32 opcode, sint32 imm, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFImmValue32);

		insn.imm = imm;
	}

	void BasicBlock::encode(uint32 opcode, const mem_ref_t& dst, const varany& src, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFWriteMemory);

		insn.srcvar1 = src.val;
		insn.membase = dst.base;
		insn.memindex = dst.index;
		insn.memscale = dst.scale;
		insn.memdisp = dst.disp;
	}

	void BasicBlock::encode(uint32 opcode, const varany& dst, const mem_ref_t& src, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFReadMemory);

		insn.dstvar1 = dst.val;
		insn.membase = src.base;
		insn.memindex = src.index;
		insn.memscale = src.scale;
		insn.memdisp = src.disp;
	}

	void BasicBlock::encode(uint32 opcode, const varany& dst, const varany& src, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFModrmReg);

		insn.dstvar1 = dst.val;
		insn.srcvar2 = src.val;
	}

	void BasicBlock::encode_imm32(uint32 opcode, const varany& dst, const varany& src, sint32 imm, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFModrmReg | kFImmValue32);

		insn.dstvar1 = dst.val;
		insn.srcvar2 = src.val;
		insn.imm = imm;
	}

	void BasicBlock::encode_imm32(uint32 opcode, const varany& dst, const mem_ref_t& src, sint32 imm, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFModrmReg | kFImmValue32 | kFReadMemory);

		insn.dstvar1 = dst.val;
		insn.membase = src.base;
		insn.memindex = src.index;
		insn.memscale = src.scale;
		insn.memdisp = src.disp;
		insn.imm = imm;
	}

	void BasicBlock::encode_w(uint32 opcode, const varany& dst, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags);

		insn.dstvar1 = dst.val;
	}

	void BasicBlock::encode_r(uint32 opcode, const varany& src, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags);

		insn.srcvar1 = src.val;
	}

	void BasicBlock::encode_r(uint32 opcode, const mem_ref_t& src, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFReadMemory);

		insn.membase = src.base;
		insn.memindex = src.index;
		insn.memscale = src.scale;
		insn.memdisp = src.disp;
	}

	void BasicBlock::encode_r_imm8(uint32 opcode, const varany& src, sint32 imm, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFImmValue8);

		insn.srcvar2 = src.val;
		insn.imm = imm;
	}

	void BasicBlock::encode_opcode_w(uint32 opcode, const varany& dst, sint32 imm, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFImmValue32 | kFOpcodeReg);

		insn.dstvar1 = dst.val;
		insn.imm = imm;
	}

	void BasicBlock::encode_rmw(uint32 opcode, const mem_ref_t& dst, const varany& src, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFReadMemory | kFWriteMemory);

		insn.srcvar1 = src.val;
		insn.membase = dst.base;
		insn.memindex = dst.index;
		insn.memscale = dst.scale;
		insn.memdisp = dst.disp;

		mInsns.push_back(insn);
	}

	void BasicBlock::encode_rmw(uint32 opcode, const varany& dst, const mem_ref_t& src, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFReadMemory);

		insn.dstvar1 = dst.val;
		insn.srcvar1 = dst.val;
		insn.membase = src.base;
		insn.memindex = src.index;
		insn.memscale = src.scale;
		insn.memdisp = src.disp;
	}

	void BasicBlock::encode_rmw(uint32 opcode, const varany& dst, sint32 imm, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFImmValue32);

		insn.dstvar2 = dst.val;
		insn.srcvar2 = dst.val;
		insn.imm = imm;
	}

	void BasicBlock::encode_rmw_imm8(uint32 opcode, const varany& dst, sint8 imm, uint32 flags) {
		Insn& insn = CreateNewInsn(opcode, flags | kFImmValue8 | kFModrmReg);

		insn.dstvar2 = dst.val;
		insn.srcvar2 = dst.val;
		insn.imm = imm;
	}

	void BasicBlock::encode_jcc(uint32 opcode, BasicBlock *target) {
		Insn& insn = CreateNewInsn(opcode, kFReadFlags | kFBranch32);

		mpTargetTrue = target;
	}

	void BasicBlock::encode_call(uint32 opcode, BasicBlock *target) {
		Insn& insn = CreateNewInsn(opcode, kFReadFlags | kFBranch32);

		mpTargetTrue = target;
	}

	void BasicBlock::encode_fpu_load(uint32 opcode, const mem_ref_t& src) {
		Insn& insn = CreateNewInsn(opcode, kFReadMemory | kFExtGroup);

		insn.membase = src.base;
		insn.memindex = src.index;
		insn.memscale = src.scale;
		insn.memdisp = src.disp;
	}

	void BasicBlock::encode_fpu_store(uint32 opcode, const varf& dst) {
		Insn& insn = CreateNewInsn(opcode, 0);

		insn.dstvar2 = dst.val;
	}

	void BasicBlock::encode_fpu_store(uint32 opcode, const mem_ref_t& dst) {
		Insn& insn = CreateNewInsn(opcode, kFWriteMemory | kFExtGroup);

		insn.membase = dst.base;
		insn.memindex = dst.index;
		insn.memscale = dst.scale;
		insn.memdisp = dst.disp;
	}

	void BasicBlock::encode_fpu_alu(uint32 opcode, const varf& src) {
		Insn& insn = CreateNewInsn(opcode, kFOpcodeReg);

		insn.srcvar1 = src.val;
	}

	void BasicBlock::encode_fpu_alu_r(uint32 opcode, const varf& dst) {
		Insn& insn = CreateNewInsn(opcode, 0);

		insn.dstvar2 = dst.val;
		insn.srcvar2 = dst.val;
	}

	void BasicBlock::encode_mmx(uint32 opcode, const varq& dst, const varq& src) {
		Insn& insn = CreateNewInsn(opcode, 0);

		insn.dstvar1 = dst.val;
		insn.srcvar2 = src.val;
	}

	void BasicBlock::encode_mmx_load(uint32 opcode, const varq& dst, const mem_ref_t& src) {
		Insn& insn = CreateNewInsn(opcode, kFReadMemory);

		insn.membase = src.base;
		insn.memindex = src.index;
		insn.memscale = src.scale;
		insn.memdisp = src.disp;
		insn.dstvar1 = dst.val;
	}

	void BasicBlock::encode_mmx_store(uint32 opcode, const mem_ref_t& dst, const varq& src) {
		Insn& insn = CreateNewInsn(opcode, kFWriteMemory);

		insn.membase = dst.base;
		insn.memindex = dst.index;
		insn.memscale = dst.scale;
		insn.memdisp = dst.disp;
		insn.srcvar1 = src.val;
	}

	void BasicBlock::encode_mmx_rmw(uint32 opcode, const varq& dst, const varq& src) {
		Insn& insn = CreateNewInsn(opcode, 0);

		insn.dstvar1 = dst.val;
		insn.srcvar1 = dst.val;
		insn.srcvar2 = src.val;
	}

	void BasicBlock::encode_mmx_rmw(uint32 opcode, const varq& dst, const mem_ref_t& src) {
		Insn& insn = CreateNewInsn(opcode, kFReadMemory);

		insn.membase = src.base;
		insn.memindex = src.index;
		insn.memscale = src.scale;
		insn.memdisp = src.disp;
		insn.dstvar1 = dst.val;
		insn.srcvar1 = dst.val;
	}

	void BasicBlock::encode_sse(uint32 opcode, const varx& dst, const var& src) {
		Insn& insn = CreateNewInsn(opcode, kFModrmReg);

		insn.dstvar1 = dst.val;
		insn.srcvar2 = src.val;
	}

	void BasicBlock::encode_sse(uint32 opcode, const varx& dst, const varx& src) {
		Insn& insn = CreateNewInsn(opcode, kFModrmReg);

		insn.dstvar1 = dst.val;
		insn.srcvar2 = src.val;
	}

	void BasicBlock::encode_sse_imm8(uint32 opcode, const varx& dst, const varx& src, uint8 imm8) {
		Insn& insn = CreateNewInsn(opcode, kFImmValue8 | kFModrmReg);

		insn.dstvar1 = dst.val;
		insn.srcvar2 = src.val;
		insn.imm = imm8;
	}

	void BasicBlock::encode_sse_load(uint32 opcode, const varx& dst, const mem_ref_t& src) {
		Insn& insn = CreateNewInsn(opcode, kFReadMemory);

		insn.membase = src.base;
		insn.memindex = src.index;
		insn.memscale = src.scale;
		insn.memdisp = src.disp;
		insn.dstvar1 = dst.val;
	}

	void BasicBlock::encode_sse_store(uint32 opcode, const mem_ref_t& dst, const varx& src) {
		Insn& insn = CreateNewInsn(opcode, kFWriteMemory);

		insn.membase = dst.base;
		insn.memindex = dst.index;
		insn.memscale = dst.scale;
		insn.memdisp = dst.disp;
		insn.srcvar1 = src.val;
	}

	void BasicBlock::encode_sse_rmw(uint32 opcode, const varx& dst, const varx& src) {
		Insn& insn = CreateNewInsn(opcode, kFModrmReg);

		insn.dstvar1 = dst.val;
		insn.srcvar1 = dst.val;
		insn.srcvar2 = src.val;
	}

	void BasicBlock::encode_sse_rmw(uint32 opcode, const varx& dst, const mem_ref_t& src) {
		Insn& insn = CreateNewInsn(opcode, kFReadMemory);

		insn.membase = src.base;
		insn.memindex = src.index;
		insn.memscale = src.scale;
		insn.memdisp = src.disp;
		insn.dstvar1 = dst.val;
		insn.srcvar1 = dst.val;
	}

	Insn& BasicBlock::CreateNewInsn(uint32 opcode, uint32 flags) {
		mInsns.push_back(Insn());

		Insn& insn = mInsns.back();

		insn.opcode = opcode;
		insn.flags = flags;
		insn.dstvar1 = 0;
		insn.dstvar2 = 0;
		insn.srcvar1 = 0;
		insn.srcvar2 = 0;
		insn.membase = 0;
		insn.memindex = 0;
		insn.memscale = 0;
		insn.memdisp = 0;
		insn.imm = 0;

		return insn;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////

	DynamicContext::DynamicContext()
	{
		mpEntryBlock = CreateBasicBlock();
		mpExitBlock = CreateBasicBlock();

		mpEntryBlock->sub(esp, 16);
		mpEntryBlock->mov(dword_ptr[esp], ebx);
		mpEntryBlock->mov(dword_ptr[esp+4], esi);
		mpEntryBlock->mov(dword_ptr[esp+8], edi);
		mpEntryBlock->mov(dword_ptr[esp+12], ebp);

		mpExitBlock->mov(ebp, dword_ptr[esp+12]);
		mpExitBlock->mov(edi, dword_ptr[esp+8]);
		mpExitBlock->mov(esi, dword_ptr[esp+4]);
		mpExitBlock->mov(ebx, dword_ptr[esp]);
		mpExitBlock->add(esp, 16);
		mpExitBlock->ret();
	}

	void DynamicContext::Copy(void *dst) {
		if (mCode.empty())
			return;

		memcpy(dst, &mCode[0], mCode.size());

		for(AbsRelocs::const_iterator it(mAbsRelocs.begin()), itEnd(mAbsRelocs.end()); it!=itEnd; ++it) {
			const AbsRelocs::value_type& reloc = *it;

			*(long *)((char *)dst + (reloc.first - 4)) = ((char *)reloc.second - (char *)dst) - reloc.first;
		}
	}

	BasicBlock *DynamicContext::CreateBasicBlock() {
		mBasicBlocks.push_back(BasicBlock());

		return &mBasicBlocks.back();
	}

	BasicBlock *DynamicContext::CreateAbsTarget(void *p) {
		mBasicBlocks.push_back(BasicBlock());
		BasicBlock& bb = mBasicBlocks.back();

		bb.mpAddress = p;

		return &bb;
	}

	void DynamicContext::Generate() {
		// allocate vars on stack
		size_t numVars = mVars.size();
		int stackoffset = 0;

		for(int i=0; i<(int)numVars; ++i) {
			static const int kSizeLookup[]={0,4,4,4,8,8,16};

			VarInfo& vi = mVars[i];

			stackoffset = ((stackoffset-1) | (kSizeLookup[vi.mSize] - 1)) + 1;
			vi.mStackOffset = stackoffset;
			stackoffset += kSizeLookup[vi.mSize];
		}

		GenerateBlocks(mpEntryBlock);

		// backpatch branches
		while(!mRelocs.empty()) {
			const std::pair<uint32, BasicBlock *>& reloc = mRelocs.back();

			*(long *)&mCode[reloc.first - 4] = reloc.second->mOffset - reloc.first;

			mRelocs.pop_back();
		}
	}

	void DynamicContext::GenerateBlocks(BasicBlock *b) {
		if (b->mOffset >= 0 || b->mpAddress)
			return;

		BasicBlock::Insns::const_iterator it(b->mInsns.begin()), itEnd(b->mInsns.end());

		b->mOffset = (sint32)mCode.size();

		for(; it!=itEnd; ++it) {
			const Insn& insn = *it;

			// emit third prefix byte
			if (insn.opcode & 0xff000000)
				mCode.push_back(0xff & (insn.opcode >> 24));

			// emit second prefix byte (66/F2/F3)
			if (insn.opcode & 0xff0000)
				mCode.push_back(0xff & (insn.opcode >> 16));

			// emit first prefix byte (0F)
			if (insn.opcode & 0x00ff00)
				mCode.push_back(0xff & (insn.opcode >> 8));

			// emit opcode
			if (insn.flags & kFOpcodeReg)
				mCode.push_back(0xff & (insn.opcode + ((insn.srcvar1 | insn.dstvar1)&7)));
			else if (!(insn.flags & kFExtGroup))
				mCode.push_back(0xff & insn.opcode);

			// emit modrm and sib bytes
			uint8 modrm = 0;
			uint8 sib = 0;

			if (insn.flags & (kFReadMemory | kFWriteMemory)) {
				if (!insn.membase && !insn.memindex) {
					modrm = 0x05;
				} else {
					if ((insn.memdisp + 0x80) & 0xFFFFFF00)
						modrm = 0x80;
					else if (insn.memdisp || (!insn.memindex && insn.membase == 5))		// [r32 + disp8] or [ebp] -> [ebp+0]
						modrm = 0x40;

					if (insn.memindex || (insn.membase&7) == 4) {		// complex or [esp+nn]
						modrm += 4;

						static const uint8 kIndexLookup[]={0,0x00,0x40,0,0x80,0,0,0,0xc0};

						if (insn.membase) {		// [base+index*scale]
							if (insn.membase == 5 && modrm < 0x40)
								modrm += 0x40;	// [ebp+index*scale] -> [ebp+index*scale+0]

							if (insn.memindex)
								sib = kIndexLookup[insn.memscale] + ((insn.memindex & 7) << 3) + (insn.membase & 7);
							else
								sib = kIndexLookup[insn.memscale] + (4 << 3) + (insn.membase & 7);
						} else {				// [index*scale] -> [index*scale + 0]
							modrm &= 0x3f;
							sib = kIndexLookup[insn.memscale] + ((insn.memindex & 7) << 3) + 5;
						}
					} else
						modrm += insn.membase & 7;
				}
			} else if (insn.srcvar2 | insn.dstvar2)
				modrm += ((insn.srcvar2 | insn.dstvar2) & 7) | 0xc0;

			if (insn.flags & kFExtGroup)
				modrm += (uint8)insn.opcode;
			else if (insn.srcvar1 | insn.dstvar1)
				modrm += ((insn.srcvar1 | insn.dstvar1) & 7) << 3;

			if (insn.flags & (kFReadMemory | kFWriteMemory | kFExtGroup | kFModrmReg)) {
				mCode.push_back(modrm);

				if (modrm < 0xc0 && (modrm&7) == 4)
					mCode.push_back(sib);
			}

			// emit displacement
			if ((modrm & 0xc0) == 0x40) {
				mCode.push_back((uint8)insn.memdisp);
			} else if ((modrm & 0xc0) == 0x80 ||
				(modrm & 0xc7) == 0x05 ||
				((modrm & 0xc7) == 0x04 && (sib & 7) == 5))
			{
				mCode.push_back(0xff & (insn.memdisp));
				mCode.push_back(0xff & (insn.memdisp >> 8));
				mCode.push_back(0xff & (insn.memdisp >> 16));
				mCode.push_back(0xff & (insn.memdisp >> 24));
			}

			// emit immediate
			if (insn.flags & kFImmValue8)
				mCode.push_back(0xff & insn.imm);
			else if (insn.flags & kFImmValue32) {
				mCode.push_back(0xff & (insn.imm));
				mCode.push_back(0xff & (insn.imm >> 8));
				mCode.push_back(0xff & (insn.imm >> 16));
				mCode.push_back(0xff & (insn.imm >> 24));
			}

			// emit branch
			if (insn.flags & kFBranch32) {
				mCode.insert(mCode.end(), 4, 0);
				AddBranch32(b->mpTargetTrue);
			}
		}

		if (b->mpTargetFalse) {
			if (b->mpTargetFalse->mOffset >= 0 || b->mpAddress) {
				mCode.push_back(0xe9);
				mCode.insert(mCode.end(), 4, 0);
				AddBranch32(b->mpTargetFalse);
			} else
				GenerateBlocks(b->mpTargetFalse);
		}

		if (b->mpTargetTrue)
			GenerateBlocks(b->mpTargetTrue);
	}

	uint32 DynamicContext::AllocVar(VarSize size) {
		mVars.push_back(VarInfo());
		mVars.back().mSize = size;

		return (uint32)mVars.size() + 0xff;
	}

	void DynamicContext::AddBranch32(BasicBlock *target) {
		if (target->mpAddress)
			mAbsRelocs.push_back(std::pair<uint32, void *>((uint32)mCode.size(), target->mpAddress));
		else
			mRelocs.push_back(std::pair<uint32, BasicBlock *>((uint32)mCode.size(), target));
	}
}
