//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef JIT_H
#define JIT_H

#ifdef _MSC_VER
	#pragma once
#endif

#include <vd2/system/vdtypes.h>

#if 0
class VDJITAllocator {
public:
	VDJITAllocator();
	~VDJITAllocator();

	void *Allocate(size_t len);
	void Free(void *p, size_t len);

	void EndUpdate(void *p, size_t len);

protected:
	void Copy(void *dst, const void *src, size_t len);

	typedef std::map<void *, size_t> FreeChunks;
	FreeChunks mFreeChunks;
	FreeChunks::iterator mNextChunk;

	typedef std::map<void *, size_t> Allocations;
	Allocations mAllocations;

	uintptr		mAllocationGranularity;
};
#endif

#include <list>
#include <vector>

namespace x86 {
	class DynamicContext;

	struct varb {
		uint32 val;
	};

	struct varw {
		uint32 val;
	};

	struct var {
		uint32 val;
	};

	struct varq {
		uint32 val;
	};

	struct varx {
		uint32 val;
	};

	struct varf {
		uint32 val;
	};

	struct mvar {
		uint32 val;
	};
	
	struct xvar {
		uint32 val;
	};

	struct var_scaled {
		uint32 val;
		int scale;
	};
	
	struct var_indexed {
		uint32 val;
		uint32 val2;
		int scale;
		sint32 disp;
	};

	inline var_scaled operator*(var a, int scale) {
		var_scaled v = {a.val, scale};
		return v;
	}
	
	inline var_indexed operator+(var a, var b) {
		var_indexed v = {a.val, b.val, 1};
		return v;
	}

	inline var_indexed operator+(var a, var_scaled b) {
		var_indexed v = {a.val, b.val, b.scale};
		return v;
	}

	inline var_indexed operator+(var_scaled b, var a) {
		var_indexed v = {a.val, b.val, b.scale};
		return v;
	}

	inline var_indexed operator+(var a, sint32 disp) {
		var_indexed v = {a.val, 0, 1, disp};
		return v;
	}

	inline var_indexed operator+(var_scaled a, sint32 disp) {
		var_indexed v = {0, a.val, a.scale, disp};
		return v;
	}

	inline var_indexed operator+(var_indexed a, sint32 disp) {
		var_indexed v = {a.val, a.val2, a.scale, a.disp + disp};
		return v;
	}

	inline var_indexed operator-(var a, sint32 disp) {
		var_indexed v = {a.val, 0, 1, -disp};
		return v;
	}

	inline var_indexed operator-(var_scaled a, sint32 disp) {
		var_indexed v = {0, a.val, a.scale, -disp};
		return v;
	}

	inline var_indexed operator-(var_indexed a, sint32 disp) {
		var_indexed v = {a.val, a.val2, a.scale, a.disp - disp};
		return v;
	}

	struct mem_ref_t {
		uint32 base;
		uint32 index;
		int scale;
		sint32 disp;
	};

	struct mem_ref_adapter_t : public mem_ref_t {
		mem_ref_adapter_t(const var& a) {
			base = a.val;
			index = 0;
			scale = 0;
			disp = 0;
		}

		mem_ref_adapter_t(const var_scaled& a) {
			base = a.val;
			index = 0;
			scale = a.scale;
			disp = 0;
		}

		mem_ref_adapter_t(const var_indexed& a) {
			base = a.val;
			index = a.val2;
			scale = a.scale;
			disp = a.disp;
		}

		mem_ref_adapter_t(const void *p) {
			base = 0;
			index = 0;
			scale = 0;
			disp = (sint32)p;
		}
	};

	struct byte_ref_t : mem_ref_t {};
	struct word_ref_t : mem_ref_t {};
	struct dword_ref_t : mem_ref_t {};
	struct qword_ref_t : mem_ref_t {};
	struct mmword_ref_t : mem_ref_t {};
	struct xmmword_ref_t : mem_ref_t {};

	struct byte_ptr_t {
		byte_ref_t operator[](const mem_ref_adapter_t& a) const {
			byte_ref_t ref;
			static_cast<mem_ref_t&>(ref) = a;
			return ref;
		}
	};
	
	struct word_ptr_t {
		word_ref_t operator[](const mem_ref_adapter_t& a) const {
			word_ref_t ref;
			static_cast<mem_ref_t&>(ref) = a;
			return ref;
		}
	};
	
	struct dword_ptr_t {
		dword_ref_t operator[](const mem_ref_adapter_t& a) const {
			dword_ref_t ref;
			static_cast<mem_ref_t&>(ref) = a;
			return ref;
		}
	};
	
	struct qword_ptr_t {
		qword_ref_t operator[](const mem_ref_adapter_t& a) const {
			qword_ref_t ref;
			static_cast<mem_ref_t&>(ref) = a;
			return ref;
		}
	};
	
	struct mmword_ptr_t {
		mmword_ref_t operator[](const mem_ref_adapter_t& a) const {
			mmword_ref_t ref;
			static_cast<mem_ref_t&>(ref) = a;
			return ref;
		}
	};

	struct xmmword_ptr_t {
		xmmword_ref_t operator[](const mem_ref_adapter_t& a) const {
			xmmword_ref_t ref;
			static_cast<mem_ref_t&>(ref) = a;
			return ref;
		}
	};
	
	const byte_ptr_t byte_ptr;
	const word_ptr_t word_ptr;
	const dword_ptr_t dword_ptr;
	const qword_ptr_t qword_ptr;
	const qword_ptr_t mmword_ptr;
	const xmmword_ptr_t xmmword_ptr;

	const varb al	= {0x10};
	const varb cl	= {0x11};
	const varb dl	= {0x12};
	const varb bl	= {0x13};
	const varb spl	= {0x14};
	const varb bpl	= {0x15};
	const varb sil	= {0x16};
	const varb dil	= {0x17};
	const varb r8b	= {0x18};
	const varb r9b	= {0x19};
	const varb r10b	= {0x1A};
	const varb r11b	= {0x1B};
	const varb r12b	= {0x1C};
	const varb r13b	= {0x1D};
	const varb r14b	= {0x1E};
	const varb r15b	= {0x1F};
	const varb ah	= {0x24};
	const varb ch	= {0x25};
	const varb dh	= {0x26};
	const varb bh	= {0x27};
	const varw ax	= {0x30};
	const varw cx	= {0x31};
	const varw dx	= {0x32};
	const varw bx	= {0x33};
	const varw sp	= {0x34};
	const varw bp	= {0x35};
	const varw si	= {0x36};
	const varw di	= {0x37};
	const varw r8w	= {0x38};
	const varw r9w	= {0x39};
	const varw r10w	= {0x3A};
	const varw r11w	= {0x3B};
	const varw r12w	= {0x3C};
	const varw r13w	= {0x3D};
	const varw r14w	= {0x3E};
	const varw r15w	= {0x3F};
	const var eax	= {0x40};
	const var ecx	= {0x41};
	const var edx	= {0x42};
	const var ebx	= {0x43};
	const var esp	= {0x44};
	const var ebp	= {0x45};
	const var esi	= {0x46};
	const var edi	= {0x47};
	const var r8d	= {0x48};
	const var r9d	= {0x49};
	const var r10d	= {0x4A};
	const var r11d	= {0x4B};
	const var r12d	= {0x4C};
	const var r13d	= {0x4D};
	const var r14d	= {0x4E};
	const var r15d	= {0x4F};

	const varq rax	= {0x50};
	const varq rdx	= {0x51};
	const varq rcx	= {0x52};
	const varq rbx	= {0x53};
	const varq rsp	= {0x54};
	const varq rbp	= {0x55};
	const varq rsi	= {0x56};
	const varq rdi	= {0x57};
	const varq r8	= {0x58};
	const varq r9	= {0x59};
	const varq r10	= {0x5A};
	const varq r11	= {0x5B};
	const varq r12	= {0x5C};
	const varq r13	= {0x5D};
	const varq r14	= {0x5E};
	const varq r15	= {0x5F};

	const varq mm0	= {0x60};
	const varq mm1	= {0x61};
	const varq mm2	= {0x62};
	const varq mm3	= {0x63};
	const varq mm4	= {0x64};
	const varq mm5	= {0x65};
	const varq mm6	= {0x66};
	const varq mm7	= {0x67};

	const varf st0	= {0x68};
	const varf st1	= {0x69};
	const varf st2	= {0x6A};
	const varf st3	= {0x6B};
	const varf st4	= {0x6C};
	const varf st5	= {0x6D};
	const varf st6	= {0x6E};
	const varf st7	= {0x6F};

	const varx xmm0		= {0x70};
	const varx xmm1		= {0x71};
	const varx xmm2		= {0x72};
	const varx xmm3		= {0x73};
	const varx xmm4		= {0x74};
	const varx xmm5		= {0x75};
	const varx xmm6		= {0x76};
	const varx xmm7		= {0x77};
	const varx xmm8		= {0x78};
	const varx xmm9		= {0x79};
	const varx xmm10	= {0x7A};
	const varx xmm11	= {0x7B};
	const varx xmm12	= {0x7C};
	const varx xmm13	= {0x7D};
	const varx xmm14	= {0x7E};
	const varx xmm15	= {0x7F};

	struct varx_generic {
		const varx operator()(int index) const {
			varx r = { 0x70 + index };

			return r;
		}
	};

	const varx_generic xmm;

	struct varf_generic {
		varf operator()(int index) const {
			varf r = { 0x68 + index };

			return r;
		}
	};

	const varf_generic st;

	struct varany {
		uint32 val;

		varany(const varb& v) : val(v.val) {}
		varany(const varw& v) : val(v.val) {}
		varany(const var& v) : val(v.val) {}
		varany(const varq& v) : val(v.val) {}
		varany(const varx& v) : val(v.val) {}
		varany(const mvar& v) : val(v.val) {}
		varany(const xvar& v) : val(v.val) {}
	};

	struct varmemany {
		uint32 val;

		varmemany(const varb& v) : val(v.val) {}
		varmemany(const varw& v) : val(v.val) {}
		varmemany(const var& v) : val(v.val) {}
		varmemany(const varq& v) : val(v.val) {}
		varmemany(const varx& v) : val(v.val) {}
		varmemany(const mvar& v) : val(v.val) {}
		varmemany(const xvar& v) : val(v.val) {}
		varmemany(const mem_ref_t& v) {}
	};

	//////////////////////////////////////////////////////////////////////////

	enum {
		kFReadFlags		= 1,
		kFWriteFlags	= 2,
		kFReadMemory	= 4,
		kFWriteMemory	= 8,
		kFImmValue8		= 16,
		kFImmValue32	= 32,
		kFBranch8		= 64,
		kFBranch32		= 128,
		kFExtGroup		= 256,	// --rrr--- in mod-r/m used for group
		kFOpcodeReg		= 512,
		kFModrmReg		= 1024
	};

	enum VarSize {
		kVarSizeB = 1,
		kVarSizeW = 2,
		kVarSizeD = 3,
		kVarSizeQ = 4,
		kVarSizeM = 5,
		kVarSizeX = 6
	};

	struct Insn {
		uint32 opcode;
		uint32 flags;
		uint32 dstvar1;		// encoded as modrm: --rrr--- or in opcode
		uint32 dstvar2;		// encoded as modrm: 11---rrr
		uint32 srcvar1;		// encoded as modrm: --rrr--- or in opcode
		uint32 srcvar2;		// encoded as modrm: 11---rrr
		uint32 membase;
		uint32 memindex;
		sint32 memscale;
		sint32 memdisp;
		sint32 imm;
	};

	class BasicBlock {
		friend class DynamicContext;
	public:
		BasicBlock();
		~BasicBlock();

		void set_address(void *p) { mpAddress = p; }

		void test(const varb& src, uint8 value) { encode_r_imm8(0xf600, src, value, kFWriteFlags | kFExtGroup); }

		void mov(const byte_ref_t& dst, const varb& src) { encode(0x88, dst, src); }
		void mov(const varb& dst, const byte_ref_t& src) { encode(0x8A, dst, src); }
		void mov(const dword_ref_t& dst, const var& src) { encode(0x89, dst, src); }
		void mov(const var& dst, const dword_ref_t& src) { encode(0x8B, dst, src); }
		void mov(const var& dst, const var& src) { encode(0x8B, dst, src); }
		void mov(const var& dst, sint32 value) { encode_opcode_w(0xB8, dst, value); }

		void lea(const var& dst, const mem_ref_t& src) { encode(0x8D, dst, src); }

		void add(const byte_ref_t& dst, const varb& src) { encode_rmw(0x00, dst, src, kFWriteFlags); }
		void add(const varb& dst, const byte_ref_t& src) { encode_rmw(0x02, dst, src, kFWriteFlags); }
		void add(const dword_ref_t& dst, const var& src) { encode_rmw(0x01, dst, src, kFWriteFlags); }
		void add(const var& dst, const dword_ref_t& src) { encode_rmw(0x03, dst, src, kFWriteFlags); }
		void add(const var& dst, const var& src) { encode(0x03, dst, src, kFWriteFlags); }
		void add(const var& dst, sint32 value) { encode_rmw(0x8100, dst, value, kFWriteFlags | kFExtGroup); }

		void or(const byte_ref_t& dst, const varb& src) { encode_rmw(0x08, dst, src, kFWriteFlags); }
		void or(const varb& dst, const byte_ref_t& src) { encode_rmw(0x0a, dst, src, kFWriteFlags); }
		void or(const dword_ref_t& dst, const var& src) { encode_rmw(0x09, dst, src, kFWriteFlags); }
		void or(const var& dst, const dword_ref_t& src) { encode_rmw(0x0b, dst, src, kFWriteFlags); }
		void or(const var& dst, const var& src) { encode(0x0b, dst, src, kFWriteFlags); }
		void or(const var& dst, sint32 value) { encode_rmw(0x8108, dst, value, kFWriteFlags | kFExtGroup); }

		void sbb(const byte_ref_t& dst, const varb& src) { encode_rmw(0x18, dst, src, kFWriteFlags); }
		void sbb(const dword_ref_t& dst, const var& src) { encode_rmw(0x19, dst, src, kFWriteFlags); }
		void sbb(const varb& dst, const byte_ref_t& src) { encode_rmw(0x1A, dst, src, kFWriteFlags); }
		void sbb(const var& dst, const dword_ref_t& src) { encode_rmw(0x1B, dst, src, kFWriteFlags); }
		void sbb(const var& dst, const var& src) { encode(0x1B, dst, src, kFWriteFlags); }
		void sbb(const var& dst, sint32 value) { encode_rmw(0x8118, dst, value, kFWriteFlags | kFExtGroup); }

		void and(const byte_ref_t& dst, const varb& src) { encode_rmw(0x20, dst, src, kFWriteFlags); }
		void and(const varb& dst, const byte_ref_t& src) { encode_rmw(0x22, dst, src, kFWriteFlags); }
		void and(const dword_ref_t& dst, const var& src) { encode_rmw(0x21, dst, src, kFWriteFlags); }
		void and(const var& dst, const dword_ref_t& src) { encode_rmw(0x23, dst, src, kFWriteFlags); }
		void and(const var& dst, const var& src) { encode(0x23, dst, src, kFWriteFlags); }
		void and(const var& dst, sint32 value) { encode_rmw(0x8120, dst, value, kFWriteFlags | kFExtGroup); }

		void sub(const byte_ref_t& dst, const varb& src) { encode_rmw(0x28, dst, src, kFWriteFlags); }
		void sub(const varb& dst, const byte_ref_t& src) { encode_rmw(0x2A, dst, src, kFWriteFlags); }
		void sub(const dword_ref_t& dst, const var& src) { encode_rmw(0x29, dst, src, kFWriteFlags); }
		void sub(const var& dst, const dword_ref_t& src) { encode_rmw(0x2B, dst, src, kFWriteFlags); }
		void sub(const var& dst, const var& src) { encode(0x2B, dst, src, kFWriteFlags); }
		void sub(const var& dst, sint32 value) { encode_rmw(0x8128, dst, value, kFWriteFlags | kFExtGroup); }

		void xor(const byte_ref_t& dst, const varb& src) { encode_rmw(0x30, dst, src, kFWriteFlags); }
		void xor(const varb& dst, const byte_ref_t& src) { encode_rmw(0x32, dst, src, kFWriteFlags); }
		void xor(const dword_ref_t& dst, const var& src) { encode_rmw(0x31, dst, src, kFWriteFlags); }
		void xor(const var& dst, const dword_ref_t& src) { encode_rmw(0x33, dst, src, kFWriteFlags); }
		void xor(const var& dst, const var& src) { encode(0x33, dst, src, kFWriteFlags); }
		void xor(const var& dst, sint32 value) { encode_rmw(0x8130, dst, value, kFWriteFlags | kFExtGroup); }

		void cmp(const byte_ref_t& dst, const varb& src) { encode_rmw(0x38, dst, src, kFWriteFlags); }
		void cmp(const dword_ref_t& dst, const var& src) { encode_rmw(0x39, dst, src, kFWriteFlags); }
		void cmp(const varb& dst, const byte_ref_t& src) { encode_rmw(0x3A, dst, src, kFWriteFlags); }
		void cmp(const var& dst, const dword_ref_t& src) { encode_rmw(0x3B, dst, src, kFWriteFlags); }
		void cmp(const var& dst, const var& src) { encode(0x3B, dst, src, kFWriteFlags); }
		void cmp(const var& dst, sint32 value) { encode_rmw(0x8138, dst, value, kFWriteFlags | kFExtGroup); }

		void cdq() { encode(0x99); }
		void movzx(const var& dst, const byte_ref_t& src) { encode(0x0FB6, dst, src); }

		void pop(const var& dst) { encode_w(0x58, dst, kFOpcodeReg); }

		void push(sint32 value) { encode_imm(0x68, value); }
		void push(const var& src) { encode_r(0x50, src, kFOpcodeReg); }
		void push(const dword_ref_t& src) { encode_r(0xFF30, src, kFExtGroup); }

		void jo (BasicBlock *target) { encode_jcc(0x0F80, target); }
		void jno(BasicBlock *target) { encode_jcc(0x0F81, target); }
		void jc (BasicBlock *target) { encode_jcc(0x0F82, target); }
		void jnc(BasicBlock *target) { encode_jcc(0x0F83, target); }
		void jz (BasicBlock *target) { encode_jcc(0x0F84, target); }
		void jnz(BasicBlock *target) { encode_jcc(0x0F85, target); }
		void jbe(BasicBlock *target) { encode_jcc(0x0F86, target); }
		void ja (BasicBlock *target) { encode_jcc(0x0F87, target); }
		void js (BasicBlock *target) { encode_jcc(0x0F88, target); }
		void jns(BasicBlock *target) { encode_jcc(0x0F89, target); }
		void jp (BasicBlock *target) { encode_jcc(0x0F8A, target); }
		void jnp(BasicBlock *target) { encode_jcc(0x0F8B, target); }
		void jl (BasicBlock *target) { encode_jcc(0x0F8C, target); }
		void jge(BasicBlock *target) { encode_jcc(0x0F8D, target); }
		void jle(BasicBlock *target) { encode_jcc(0x0F8E, target); }
		void jg (BasicBlock *target) { encode_jcc(0x0F8F, target); }
		
		void jmp(BasicBlock *target) {
			mpTargetFalse = target;
		}

		void call(BasicBlock *target) {
			encode_call(0xE8, target);
		}

		void shl(const var& dst, int count) { encode_rmw_imm8(0xC120, dst, (sint8)count, kFExtGroup); }

		void imul(const var& dst, const var& src, sint32 value) { encode_imm32(0x69, dst, src, value); }
		void imul(const var& dst, const mem_ref_t& src, sint32 value) { encode_imm32(0x69, dst, src, value); }

		void ret() { encode(0xc3); }

		// FPU INSTRUCTIONS
		void fnstsw_ax() { encode(0xDFE0); }

		void fcom() { encode(0xD8D1); }
		void fabs() { encode(0xD9E1); }
		void fld1() { encode(0xD9E8); }
		void fldz() { encode(0xD9EE); }
		void fld(const dword_ref_t& src) { encode_fpu_load(0xD900, src); }
		void fild(const dword_ref_t& src) { encode_fpu_load(0xDB00, src); }
		void fst(const dword_ref_t& dst) { encode_fpu_store(0xD910, dst); }
		void fstp(const dword_ref_t& dst) { encode_fpu_store(0xD918, dst); }
		void fstp(const varf& dst) { encode_fpu_store(0xDDD8, dst); }
		void fchs() { encode(0xD9E0); }
		void fsqrt() { encode(0xD9FA); }
		void fxch(const varf& src) { encode_fpu_alu(0xD9C8, src); }
		void fadd(const dword_ref_t& src) { encode_fpu_load(0xD800, src); }
		void fadd(const varf& src) { encode_fpu_alu(0xD8C0, src); }
		void faddr(const varf& src) { encode_fpu_alu(0xDCC0, src); }
		void faddrp(const varf& dst) { encode_fpu_alu_r(0xDEC0, dst); }
		void faddp() { encode(0xDEC1); }
		void fmul(const varf& src) { encode_fpu_alu(0xD8C8, src); }
		void fmul(const dword_ref_t& src) { encode_fpu_load(0xD808, src); }
		void fmulr(const varf& src) { encode_fpu_alu(0xDCC8, src); }
		void fmulrp(const varf& dst) { encode_fpu_alu_r(0xDEC8, dst); }
		void fmulp() { encode(0xDEC9); }
		void fsub(const varf& src) { encode_fpu_alu(0xD8E0, src); }
		void fsubr(const varf& src) { encode_fpu_alu(0xDCE8, src); }
		void fsubrp(const varf& dst) { encode_fpu_alu_r(0xDEE8, dst); }
		void fsubp() { encode(0xDEE9); }
		void fsubrp() { encode(0xDEE1); }
		void fdiv(const varf& src) { encode_fpu_alu(0xD8F8, src); }
		void fdivr(const varf& src) { encode_fpu_alu(0xDCF0, src); }
		void fdivrp(const varf& dst) { encode_fpu_alu_r(0xDEF0, dst); }
		void fdivrp() { encode(0xDEF1); }

		// MMX INSTRUCTIONS
		void movd	(const varq& dst, const mem_ref_t& src) { encode_mmx_load(0x0F6E, dst, src); }
		void movd	(const mem_ref_t& dst, const varq& src) { encode_mmx_store(0x0F7E, dst, src); }
		void movq	(const varq& dst, const mem_ref_t& src) { encode_mmx_load(0x0F6F, dst, src); }
		void movq	(const mem_ref_t& dst, const varq& src) { encode_mmx_store(0x0F7F, dst, src); }
		void movq	(const varq& dst, const varq& src) { encode_mmx(0x0F6F, dst, src); }

		void pmullw	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FD5, dst, src); }
		void pmulhw	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FE5, dst, src); }
		void pmaddwd(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FF5, dst, src); }

		void psubusb(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FD8, dst, src); }
		void psubusw(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FD9, dst, src); }
		void psubsb	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FE8, dst, src); }
		void psubsw	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FE9, dst, src); }
		void paddusb(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FDC, dst, src); }
		void paddusw(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FDD, dst, src); }
		void paddsb	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FEC, dst, src); }
		void paddsw	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FED, dst, src); }

		void pand	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FDB, dst, src); }
		void pandn	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FDF, dst, src); }
		void por	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FEB, dst, src); }
		void pxor	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FEF, dst, src); }

		void psubb	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FF8, dst, src); }
		void psubw	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FF9, dst, src); }
		void psubd	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FFA, dst, src); }
		void paddb	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FFC, dst, src); }
		void paddw	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FFD, dst, src); }
		void paddd	(const varq& dst, const varq& src) { encode_mmx_rmw(0x0FFE, dst, src); }

		// SSE instructions
		void movups	(const varx& dst, const varx& src)		{ encode_sse(0x0F10, dst, src); }
		void movups	(const varx& dst, const mem_ref_t& src)	{ encode_sse_load(0x0F10, dst, src); }
		void movups	(const mem_ref_t& dst, const varx& src)	{ encode_sse_store(0x0F11, dst, src); }
		void movss	(const varx& dst, const varx& src)		{ encode_sse(0xF30F10, dst, src); }
		void movss	(const varx& dst, const mem_ref_t& src)	{ encode_sse_load(0xF30F10, dst, src); }
		void movss	(const mem_ref_t& dst, const varx& src)	{ encode_sse_store(0xF30F11, dst, src); }
		void movaps	(const varx& dst, const varx& src)		{ encode_sse(0x0F28, dst, src); }
		void movaps	(const varx& dst, const mem_ref_t& src)	{ encode_sse_load(0x0F28, dst, src); }
		void movaps	(const mem_ref_t& dst, const varx& src)	{ encode_sse_store(0x0F29, dst, src); }
		void movlps	(const varx& dst, const varx& src)		{ encode_sse(0x0F12, dst, src); }
		void movlps	(const varx& dst, const mem_ref_t& src)	{ encode_sse_load(0x0F12, dst, src); }
		void movlps	(const mem_ref_t& dst, const varx& src)	{ encode_sse_store(0x0F13, dst, src); }
		void movhps	(const varx& dst, const varx& src)		{ encode_sse(0x0F16, dst, src); }
		void movhps	(const varx& dst, const mem_ref_t& src)	{ encode_sse_load(0x0F16, dst, src); }
		void movhps	(const mem_ref_t& dst, const varx& src)	{ encode_sse_store(0x0F17, dst, src); }
		void andps	(const varx& dst, const varx& src)		{ encode_sse_rmw(0x0F54, dst, src); }
		void andps	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x0F54, dst, src); }
		void andnps	(const varx& dst, const varx& src)		{ encode_sse_rmw(0x0F55, dst, src); }
		void andnps	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x0F55, dst, src); }
		void orps	(const varx& dst, const varx& src)		{ encode_sse_rmw(0x0F56, dst, src); }
		void orps	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x0F56, dst, src); }
		void xorps	(const varx& dst, const varx& src)		{ encode_sse_rmw(0x0F57, dst, src); }
		void xorps	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x0F57, dst, src); }
		void addps	(const varx& dst, const varx& src)		{ encode_sse_rmw(0x0F58, dst, src); }
		void addps	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x0F58, dst, src); }
		void addss	(const varx& dst, const varx& src)		{ encode_sse_rmw(0xF30F58, dst, src); }
		void addss	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0xF30F58, dst, src); }
		void mulps	(const varx& dst, const varx& src)		{ encode_sse_rmw(0x0F59, dst, src); }
		void mulps	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x0F59, dst, src); }
		void mulss	(const varx& dst, const varx& src)		{ encode_sse_rmw(0xF30F59, dst, src); }
		void mulss	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0xF30F59, dst, src); }
		void subps	(const varx& dst, const varx& src)		{ encode_sse_rmw(0x0F5C, dst, src); }
		void subps	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x0F5C, dst, src); }
		void subss	(const varx& dst, const varx& src)		{ encode_sse_rmw(0xF30F5C, dst, src); }
		void subss	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0xF30F5C, dst, src); }
		void rsqrtps(const varx& dst, const mem_ref_t& src) { encode_sse_load(0x0F52, dst, src); }
		void rsqrtps(const varx& dst, const varx& src)		{ encode_sse(0x0F52, dst, src); }
		void rsqrtss(const varx& dst, const mem_ref_t& src) { encode_sse_load(0xF30F52, dst, src); }
		void rsqrtss(const varx& dst, const varx& src)		{ encode_sse(0xF30F52, dst, src); }
		void rcpps	(const varx& dst, const mem_ref_t& src) { encode_sse_load(0x0F53, dst, src); }
		void rcpps	(const varx& dst, const varx& src)		{ encode_sse(0x0F53, dst, src); }
		void rcpss	(const varx& dst, const mem_ref_t& src) { encode_sse_load(0xF30F53, dst, src); }
		void rcpss	(const varx& dst, const varx& src)		{ encode_sse(0xF30F53, dst, src); }

		void minps	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x0F5D, dst, src); }
		void minps	(const varx& dst, const varx& src)		{ encode_sse_rmw(0x0F5D, dst, src); }
		void minss	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0xF30F5D, dst, src); }
		void minss	(const varx& dst, const varx& src)		{ encode_sse_rmw(0xF30F5D, dst, src); }
		void maxps	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x0F5F, dst, src); }
		void maxps	(const varx& dst, const varx& src)		{ encode_sse_rmw(0x0F5F, dst, src); }
		void maxss	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0xF30F5F, dst, src); }
		void maxss	(const varx& dst, const varx& src)		{ encode_sse_rmw(0xF30F5F, dst, src); }

		void cmpltps(const varx& dst, const varx& src)		{ encode_sse_imm8(0x0FC2, dst, src, 1); }

		// SSE2 instructions
		void movd	(const varx& dst, const var& src) { encode_sse(0x660F6E, dst, src); }
		void movd	(const varx& dst, const mem_ref_t& src) { encode_sse_load(0x660F6E, dst, src); }
		void movd	(const mem_ref_t& dst, const varx& src) { encode_sse_store(0x660F7E, dst, src); }
		void movq	(const varx& dst, const mem_ref_t& src) { encode_sse_load(0xF30F7E, dst, src); }
		void movq	(const mem_ref_t& dst, const varx& src) { encode_sse_store(0x660FD6, dst, src); }
		void movq	(const varx& dst, const varx& src)		{ encode_sse(0x660FD6, dst, src); }
		void movdqa	(const varx& dst, const mem_ref_t& src) { encode_sse_load(0x660F6F, dst, src); }
		void movdqa	(const mem_ref_t& dst, const varx& src) { encode_sse_store(0x660F7F, dst, src); }
		void movdqa	(const varx& dst, const varx& src)		{ encode_sse(0x660F6F, dst, src); }

		void pmullw	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FD5, dst, src); }
		void pmulhw	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FE5, dst, src); }
		void pmaddwd(const varx& dst, const varx& src) { encode_sse_rmw(0x660FF5, dst, src); }

		void psubusb(const varx& dst, const varx& src) { encode_sse_rmw(0x660FD8, dst, src); }
		void psubusw(const varx& dst, const varx& src) { encode_sse_rmw(0x660FD9, dst, src); }
		void psubsb	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FE8, dst, src); }
		void psubsw	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FE9, dst, src); }
		void paddusb(const varx& dst, const varx& src) { encode_sse_rmw(0x660FDC, dst, src); }
		void paddusw(const varx& dst, const varx& src) { encode_sse_rmw(0x660FDD, dst, src); }
		void paddsb	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FEC, dst, src); }
		void paddsw	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FED, dst, src); }

		void pand	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FDB, dst, src); }
		void pand	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x660FDB, dst, src); }
		void pandn	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FDF, dst, src); }
		void por	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FEB, dst, src); }
		void pxor	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FEF, dst, src); }

		void pcmpgtd(const varx& dst, const varx& src) { encode_sse_rmw(0x660F66, dst, src); }
		void pcmpgtd(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x660F66, dst, src); }

		void psubb	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FF8, dst, src); }
		void psubw	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FF9, dst, src); }
		void psubd	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FFA, dst, src); }
		void psubd	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x660FFA, dst, src); }
		void paddb	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FFC, dst, src); }
		void paddw	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FFD, dst, src); }
		void paddd	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FFE, dst, src); }
		void paddd	(const varx& dst, const mem_ref_t& src) { encode_sse_rmw(0x660FFE, dst, src); }
		void pavgb	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FE0, dst, src); }
		void pavgw	(const varx& dst, const varx& src) { encode_sse_rmw(0x660FE3, dst, src); }

		void psrlw	(const varx& dst, uint8 imm) { encode_r_imm8(0x660F7110, dst, imm, kFExtGroup); }
		void psraw	(const varx& dst, uint8 imm) { encode_r_imm8(0x660F7120, dst, imm, kFExtGroup); }
		void psllw	(const varx& dst, uint8 imm) { encode_r_imm8(0x660F7130, dst, imm, kFExtGroup); }
		void psrld	(const varx& dst, uint8 imm) { encode_r_imm8(0x660F7210, dst, imm, kFExtGroup); }
		void psrad	(const varx& dst, uint8 imm) { encode_r_imm8(0x660F7220, dst, imm, kFExtGroup); }
		void pslld	(const varx& dst, uint8 imm) { encode_r_imm8(0x660F7230, dst, imm, kFExtGroup); }
		void psrlq	(const varx& dst, uint8 imm) { encode_r_imm8(0x660F7310, dst, imm, kFExtGroup); }
		void psllq	(const varx& dst, uint8 imm) { encode_r_imm8(0x660F7330, dst, imm, kFExtGroup); }

		void packuswb(const varx& dst, const varx& src) { encode_sse_rmw(0x660F67, dst, src); }
		void packssdw(const varx& dst, const varx& src) { encode_sse_rmw(0x660F6B, dst, src); }

		void punpcklbw(const varx& dst, const varx& src) { encode_sse_rmw(0x660F60, dst, src); }
		void punpcklwd(const varx& dst, const varx& src) { encode_sse_rmw(0x660F61, dst, src); }

		void cvtdq2ps(const varx& dst, const varx& src) { encode_sse(0x0F5B, dst, src); }
		void cvtps2dq(const varx& dst, const varx& src) { encode_sse(0x660F5B, dst, src); }
		void cvttps2dq(const varx& dst, const varx& src) { encode_sse(0xF30F5B, dst, src); }
		void cvttps2dq(const varx& dst, const mem_ref_t& src) { encode_sse_load(0xF30F5B, dst, src); }

		void pshuflw(const varx& dst, const varx& src, uint8 imm) { encode_sse_imm8(0xF20F70, dst, src, imm); }
		void pshufhw(const varx& dst, const varx& src, uint8 imm) { encode_sse_imm8(0xF30F70, dst, src, imm); }
		void pshufd(const varx& dst, const varx& src, uint8 imm) { encode_sse_imm8(0x660F70, dst, src, imm); }

	protected:
		void encode(uint32 opcode, uint32 flags = 0);
		void encode_imm(uint32 opcode, sint32 imm, uint32 flags = 0);
		void encode(uint32 opcode, const mem_ref_t& dst, const varany& src, uint32 flags = 0);
		void encode(uint32 opcode, const varany& dst, const mem_ref_t& src, uint32 flags = 0);
		void encode(uint32 opcode, const varany& dst, const varany& src, uint32 flags = 0);
		void encode_imm32(uint32 opcode, const varany& dst, const varany& src, sint32 imm, uint32 flags = 0);
		void encode_imm32(uint32 opcode, const varany& dst, const mem_ref_t& src, sint32 imm, uint32 flags = 0);
		void encode_w(uint32 opcode, const varany& src, uint32 flags = 0);
		void encode_r(uint32 opcode, const varany& src, uint32 flags = 0);
		void encode_r(uint32 opcode, const mem_ref_t& src, uint32 flags = 0);
		void encode_r_imm8(uint32 opcode, const varany& src, sint32 imm, uint32 flags = 0);
		void encode_opcode_w(uint32 opcode, const varany& dst, sint32 imm, uint32 flags = 0);
		void encode_rmw(uint32 opcode, const mem_ref_t& dst, const varany& src, uint32 flags = 0);
		void encode_rmw(uint32 opcode, const varany& dst, const mem_ref_t& src, uint32 flags = 0);
		void encode_rmw(uint32 opcode, const varany& dst, sint32 imm, uint32 flags = 0);
		void encode_rmw_imm8(uint32 opcode, const varany& dst, sint8 imm, uint32 flags = 0);
		void encode_jcc(uint32 opcode, BasicBlock *target);
		void encode_call(uint32 opcode, BasicBlock *target);

		void encode_fpu_load(uint32 opcode, const mem_ref_t& src);
		void encode_fpu_store(uint32 opcode, const mem_ref_t& dst);
		void encode_fpu_store(uint32 opcode, const varf& dst);
		void encode_fpu_alu(uint32 opcode, const varf& src);
		void encode_fpu_alu_r(uint32 opcode, const varf& dst);

		void encode_mmx(uint32 opcode, const varq& dst, const varq& src);
		void encode_mmx_load(uint32 opcode, const varq& dst, const mem_ref_t& src);
		void encode_mmx_store(uint32 opcode, const mem_ref_t& dst, const varq& src);
		void encode_mmx_rmw(uint32 opcode, const varq& dst, const varq& src);
		void encode_mmx_rmw(uint32 opcode, const varq& dst, const mem_ref_t& src);

		void encode_sse(uint32 opcode, const varx& dst, const var& src);
		void encode_sse(uint32 opcode, const varx& dst, const varx& src);
		void encode_sse_imm8(uint32 opcode, const varx& dst, const varx& src, uint8 imm8);
		void encode_sse_load(uint32 opcode, const varx& dst, const mem_ref_t& src);
		void encode_sse_store(uint32 opcode, const mem_ref_t& dst, const varx& src);
		void encode_sse_rmw(uint32 opcode, const varx& dst, const varx& src);
		void encode_sse_rmw(uint32 opcode, const varx& dst, const mem_ref_t& src);

		Insn& CreateNewInsn(uint32 opcode, uint32 flags);

	protected:
		void *mpAddress;
		sint32 mOffset;
		typedef std::list<Insn> Insns;
		Insns mInsns;
		BasicBlock *mpTargetFalse;
		BasicBlock *mpTargetTrue;

		std::vector<uint8> mGenVars;
		std::vector<uint8> mKillVars;
	};

	class DynamicContext {
	public:
		DynamicContext();

		void Copy(void *dst);

		const void *GetCode() const { return mCode.empty() ? NULL : &mCode[0]; }
		uint32 GetCodeLen() const  { return mCode.size(); }

		BasicBlock *CreateBasicBlock();
		BasicBlock *CreateAbsTarget(void *p);
		BasicBlock *GetEntryBlock() { return mpEntryBlock; }
		BasicBlock *GetExitBlock() { return mpExitBlock; }

		varb AllocVarB() { varb v = {AllocVar(kVarSizeB)}; return v; }
		varw AllocVarW() { varw v = {AllocVar(kVarSizeW)}; return v; }
		var  AllocVarD() { var  v = {AllocVar(kVarSizeD)}; return v; }
		varq AllocVarQ() { varq v = {AllocVar(kVarSizeQ)}; return v; }
		mvar AllocVarM() { mvar v = {AllocVar(kVarSizeM)}; return v; }
		xvar AllocVarX() { xvar v = {AllocVar(kVarSizeX)}; return v; }

		void Generate();

	protected:
		void GenerateBlocks(BasicBlock *b);

		uint32 AllocVar(VarSize size);
		void AddBranch32(BasicBlock *target);

		typedef std::list<BasicBlock> BasicBlocks;
		BasicBlocks mBasicBlocks;
		BasicBlock *mpEntryBlock;
		BasicBlock *mpExitBlock;

		struct VarInfo {
			VarSize mSize;
			int mStackOffset;
		};

		std::vector<VarInfo> mVars;

		std::vector<uint8> mCode;
		std::vector<std::pair<uint32, BasicBlock *> > mRelocs;
		typedef std::vector<std::pair<uint32, void *> > AbsRelocs;
		AbsRelocs mAbsRelocs;
	};
}

#endif
