//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef LEXER_H
#define LEXER_H

#ifdef _MSC_VER
	#pragma once
#endif

class IVDShaderLogOutput;

namespace nsVDShaderTokens {
	enum VDShaderToken {
		kTokenInt		= 0x100,
		kTokenFloat,
		kTokenIdent,
		kTokenString,
		kTokenLE,
		kTokenGE,
		kTokenEQ,
		kTokenNE,
		kTokenAnd,
		kTokenOr,
		kTokenAddEq,
		kTokenSubEq,
		kTokenMulEq,
		kTokenDivEq,
	};
}

struct VDShaderErrorInfo {
	int mLine;
	int mColumn;
	int mLength;
};

class IVDShaderErrorOutput {
public:
	virtual IVDShaderLogOutput *GetLogOutput() const = 0;
	virtual bool SetError(const char *s) = 0;
	virtual bool SetErrorF(const char *format, ...) = 0;
	virtual bool SetErrorVF(const char *format, va_list val) = 0;
	virtual bool IsErrorFlagSet() const = 0;
	virtual void ClearErrors() = 0;
};

class VDShaderLexer : public IVDShaderErrorOutput {
public:
	VDShaderLexer();

	void Init(const char *src, size_t len, const char *context, IVDShaderLogOutput *pOutput);
	bool GetErrorInfo(VDShaderErrorInfo& errorInfo);

	IVDShaderLogOutput *GetLogOutput() const { return mpOutput; }
	bool IsErrorFlagSet() const { return mbError; }

	void SetReturnWhitespace(bool enable) { mbReturnWhitespace = enable; }

	int Token();
	void Push();

	bool SetError(const char *s);
	bool SetErrorF(const char *format, ...);
	bool SetErrorVF(const char *format, va_list val);
	void ClearErrors();

public:
	const char *mpToken;
	int mTokenLength;
	union {
		int		mIntVal;
		float	mFloatVal;
	};

protected:
	void SkipComment();

	const char *mpTokenLineStart;
	const char *mpSrc;
	const char *mpSrcEnd;
	const char *mpSrcLineStart;
	const char *mpContext;
	int mLineNo;

	bool mbReturnWhitespace;
	bool mbError;
	VDShaderErrorInfo mErrorInfo;

	IVDShaderLogOutput *mpOutput;
};

#endif
