#include "stdafx.h"
#include <new>
#include <windows.h>
#include <commctrl.h>
#include <vd2/system/registry.h>
#include <vd2/system/math.h>
#include <vd2/system/strutil.h>
#include <vd2/system/VDString.h>
#include <vd2/system/vdalloc.h>
#include <vd2/system/w32assist.h>
#include <vd2/plugin/vdvideofilt.h>
#include <vd2/VDXFrame/VideoFilterDialog.h>
#include "paramdialog.h"
#include "resource.h"
#include "effect.h"
#include "editor.h"

#ifdef _MSC_VER
	#pragma warning(disable: 4355)		// warning C4355: 'this' : used in base member initializer list
	#pragma comment(lib, "comctl32")
#endif

class VDShaderFilterDialog : public VDXVideoFilterDialog, public IVDShaderFilterConfigDialog {
public:
	VDShaderFilterDialog(VDShaderFilterConfig& cfg, IVDXFilterPreview *ifp, IVDShaderFilter *pFilter)
		: mConfig(cfg)
		, mifp(ifp)
		, mpFilter(pFilter)
	{
	}

	bool Show(HWND hwndParent);
	HWND ShowModeless(HWND hwndParent);

public:
	VDXHWND GetHwnd() const { return (VDXHWND)mhdlg; }

protected:
	INT_PTR DlgProc(UINT msg, WPARAM wParam, LPARAM lParam);
	void OnInit();
	void OnSize();
	void UpdateParamsUI();
	void UpdateList();
	void UpdateEnables();
	void UpdateParameters();

	enum {
		kControlBase			= 10000,
		kControlIdWidth			= 16,
		kControlIdWidthBits		= 4,
		kControlLabelOffset		= 0,
		kControlTrackbarOffset	= 1
	};

	VDShaderFilterConfig& mConfig;
	IVDXFilterPreview *mifp;
	IVDShaderFilter *mpFilter;

	int		mLabelX;
	int		mControlX1;
	int		mControlX2;
	int		mPreviewButtonY;
	int		mParamUIHeight;
	int		mParamUISpacerHeight;

	typedef std::vector<VDStringW> Files;
	Files mFiles;

	struct ParameterUIEntry {
		VDStringA	mName;
		VDFXParamType	mType;
		HWND		mhwndLabel;
		HWND		mhwndGadget;
		int			mSteps;
		VDShaderConstant	mValue;
		VDShaderConstant	mMinValue;
		VDShaderConstant	mMaxValue;
	};

	typedef std::vector<ParameterUIEntry> ParameterUI;
	ParameterUI mParameterUI;
};

bool VDShaderFilterDialog::Show(HWND hwndParent) {
	VDShaderFilterConfig oldConfig(mConfig);

	if (!VDXVideoFilterDialog::Show(VDGetLocalModuleHandleW32(), MAKEINTRESOURCE(IDD_CUSTOM_CONFIGURE), hwndParent)) {
		mConfig = oldConfig;
		return false;
	}

	return true;
}

HWND VDShaderFilterDialog::ShowModeless(HWND hwndParent) {
	if (!VDXVideoFilterDialog::ShowModeless(VDGetLocalModuleHandleW32(), MAKEINTRESOURCE(IDD_CUSTOM_CONFIGURE_CHILD), hwndParent))
		return false;

	return mhdlg;
}

INT_PTR VDShaderFilterDialog::DlgProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_INITDIALOG:
			OnInit();
			return FALSE;

		case WM_SIZE:
			OnSize();
			break;

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
			case IDC_PREVIEW:
				if (mifp)
					mifp->Toggle((VDXHWND)mhdlg);
				break;

			case IDC_KEEPSIZE:
				{
					bool checked = IsDlgButtonChecked(mhdlg, IDC_KEEPSIZE) != 0;

					if (checked != mConfig.mbKeepSize) {
						mConfig.mbKeepSize = checked;

						UpdateEnables();

						if (mifp)
							mifp->RedoSystem();
					}
				}
				break;

			case IDC_WIDTH:
				if (HIWORD(wParam) == EN_KILLFOCUS) {
					BOOL valid;
					UINT val = GetDlgItemInt(mhdlg, IDC_WIDTH, &valid, FALSE);

					if (valid) {
						mConfig.mNewWidth = val;

						if (mifp)
							mifp->RedoSystem();
					}
				}
				break;

			case IDC_HEIGHT:
				if (HIWORD(wParam) == EN_KILLFOCUS) {
					BOOL valid;
					UINT val = GetDlgItemInt(mhdlg, IDC_HEIGHT, &valid, FALSE);

					if (valid) {
						mConfig.mNewHeight = val;

						if (mifp)
							mifp->RedoSystem();
					}
				}
				break;

			case IDOK:
				EndDialog(mhdlg, true);
				return TRUE;
			case IDCANCEL:
				EndDialog(mhdlg, false);
				return TRUE;
			}
			break;

		case WM_HSCROLL:
			if (lParam) {
				HWND hwndScroll = (HWND)lParam;
				UINT id = GetWindowLong(hwndScroll, GWL_ID);
				unsigned index = (id - kControlBase) >> kControlIdWidthBits;

				if (index < mParameterUI.size()) {
					ParameterUIEntry& ent = mParameterUI[index];
					int type = (id - kControlBase) & (kControlIdWidth - 1);

					if (type == kControlTrackbarOffset) {
						int pos = SendMessage(hwndScroll, TBM_GETPOS, 0, 0);
						float v = ent.mMinValue.v[0] + (ent.mMaxValue.v[0] - ent.mMinValue.v[0]) / (float)ent.mSteps * (float)pos;

						if (ent.mMinValue.v[0] <= ent.mMaxValue.v[0]) {
							if (!(v <= ent.mMaxValue.v[0]))
								v = ent.mMaxValue.v[0];

							if (!(v >= ent.mMinValue.v[0]))
								v = ent.mMinValue.v[0];
						} else {
							if (!(v >= ent.mMaxValue.v[0]))
								v = ent.mMaxValue.v[0];

							if (!(v <= ent.mMinValue.v[0]))
								v = ent.mMinValue.v[0];
						}

						if (ent.mValue.v[0] != v) {
							ent.mValue.v[0] = v;

							UpdateParameters();
						}
					}
				}
			}
			break;
	}

	return FALSE;
}

void VDShaderFilterDialog::OnInit() {
	mPreviewButtonY = 0;

	mLabelX = 16;
	HWND hwndPreview = GetDlgItem(mhdlg, IDC_PREVIEW);
	if (hwndPreview) {
		RECT r;
		GetWindowRect(hwndPreview, &r);
		MapWindowPoints(NULL, mhdlg, (LPPOINT)&r, 2);
		mLabelX = r.left;
		mPreviewButtonY = r.top;
	}

	mControlX1 = 0;
	mControlX2 = 0;
	HWND hwndWidth = GetDlgItem(mhdlg, IDC_WIDTH);
	if (hwndWidth) {
		RECT r;
		GetWindowRect(hwndWidth, &r);
		MapWindowPoints(NULL, mhdlg, (LPPOINT)&r, 2);
		mControlX1 = r.left;
		mControlX2 = r.right;
	}

	RECT r = {0, 0, 14, 14};
	MapDialogRect(mhdlg, &r);
	mParamUIHeight = r.bottom - r.top;

	RECT r2 = {0, 0, 4, 4};
	MapDialogRect(mhdlg, &r2);
	mParamUISpacerHeight = r2.bottom - r2.top;

	bool isChild = (GetWindowLong(mhdlg, GWL_STYLE) & WS_CHILD) != 0;
	if (isChild) {
		if (hwndPreview) {
			EnableWindow(hwndPreview, FALSE);
			ShowWindow(hwndPreview, SW_HIDE);
		}
	}

	if (mConfig.mbKeepSize)
		CheckDlgButton(mhdlg, IDC_KEEPSIZE, BST_CHECKED);
	else {
		SetDlgItemInt(mhdlg, IDC_WIDTH, mConfig.mNewWidth, FALSE);
		SetDlgItemInt(mhdlg, IDC_HEIGHT, mConfig.mNewHeight, FALSE);
	}

	UpdateEnables();

	SetFocus(GetDlgItem(mhdlg, IDC_WIDTH));

	if (mifp)
		mifp->InitButton((VDXHWND)GetDlgItem(mhdlg, IDC_PREVIEW));

	UpdateParamsUI();
}

void VDShaderFilterDialog::OnSize() {
	if (!(GetWindowLong(mhdlg, GWL_STYLE) & WS_CHILD))
		return;

	vdfastvector<HWND> windows;

	windows.push_back(GetDlgItem(mhdlg, IDC_WIDTH));
	windows.push_back(GetDlgItem(mhdlg, IDC_HEIGHT));

	for(ParameterUI::const_iterator it(mParameterUI.begin()), itEnd(mParameterUI.end()); it != itEnd; ++it) {
		const ParameterUIEntry& pui = *it;

		if (pui.mhwndGadget)
			windows.push_back(pui.mhwndGadget);
	}

	RECT rc;
	GetClientRect(mhdlg, &rc);

	RECT rPad = {0,0,7,7};
	MapDialogRect(mhdlg, &rPad);

	HDWP hdwp = BeginDeferWindowPos(1);
	RECT r;
	int n = windows.size();
	for(int i=0; i<n; ++i) {
		HWND hwndControl = windows[i];
		if (!hwndControl)
			continue;

		GetWindowRect(hwndControl, &r);
		MapWindowPoints(NULL, mhdlg, (LPPOINT)&r, 2);

		const int h = r.bottom - r.top;
		int w = rc.right - (rPad.right - rPad.left) - r.left;

		if (w < 0)
			w = 0;

		HDWP hdwpNew = DeferWindowPos(hdwp, hwndControl, NULL, 0, 0, w, h, SWP_NOMOVE|SWP_NOZORDER|SWP_NOACTIVATE);
		if (!hdwpNew)
			break;
		hdwp = hdwpNew;
	}

	EndDeferWindowPos(hdwp);	
}

void VDShaderFilterDialog::UpdateParamsUI() {
	// destroy existing parameter UI
	while(!mParameterUI.empty()) {
		ParameterUIEntry& ent = mParameterUI.back();

		if (ent.mhwndLabel)
			DestroyWindow(ent.mhwndLabel);
		if (ent.mhwndGadget)
			DestroyWindow(ent.mhwndGadget);

		mParameterUI.pop_back();
	}

	typedef vdfastvector<VDFXUIBinding> ParamBindings;
	ParamBindings paramBindings;
	if (mpFilter)
		mpFilter->GetParamBindings(paramBindings);

	HFONT hfont = (HFONT)SendMessage(mhdlg, WM_GETFONT, 0, 0);

	int y = mPreviewButtonY;
	UINT id = kControlBase;
	int numConfigParams = mConfig.mParamValues.size();

	for(ParamBindings::const_iterator it(paramBindings.begin()), itEnd(paramBindings.end()); it != itEnd; ++it) {
		const VDFXUIBinding& uibinding = *it;

		mParameterUI.push_back(ParameterUIEntry());
		ParameterUIEntry& ent = mParameterUI.back();
		ent.mName = uibinding.mpName;
		ent.mType = uibinding.mParamType;
		ent.mValue = uibinding.mDefaultValue;
		ent.mMinValue = uibinding.mMinValue;
		ent.mMaxValue = uibinding.mMaxValue;

		for(int i=0; i<numConfigParams; ++i) {
			VDShaderFilterParameterConfig& paramConfig = mConfig.mParamValues[i];

			if (ent.mName == paramConfig.mName) {
				if (ent.mType == paramConfig.mType)
					ent.mValue = paramConfig.mValue;

				break;
			}
		}

		ent.mhwndLabel = CreateWindow("STATIC", uibinding.mpName, WS_CHILD|WS_VISIBLE|SS_CENTERIMAGE, mLabelX, y, mControlX1 - mLabelX, mParamUIHeight, mhdlg, (HMENU)(id + kControlLabelOffset), VDGetLocalModuleHandleW32(), NULL);
		if (ent.mhwndLabel) {
			if (hfont)
				SendMessage(ent.mhwndLabel, WM_SETFONT, (WPARAM)hfont, TRUE);
		}

		ent.mhwndGadget = CreateWindow(TRACKBAR_CLASS, "", WS_CHILD|WS_VISIBLE|TBS_NOTICKS|TBS_BOTH, mControlX1, y, mControlX2 - mControlX1, mParamUIHeight, mhdlg, (HMENU)(id + kControlTrackbarOffset), VDGetLocalModuleHandleW32(), NULL);
		if (ent.mhwndGadget) {
			if (hfont)
				SendMessage(ent.mhwndGadget, WM_SETFONT, (WPARAM)hfont, TRUE);

			if (ent.mMinValue.v[0] < -1e+10f)
				ent.mMinValue.v[0] = 0.0f;
			if (ent.mMaxValue.v[0] > 1e+10f)
				ent.mMaxValue.v[0] = 1.0f;

			ent.mSteps = uibinding.mSteps;
			if (ent.mSteps > 10000)
				ent.mSteps = 10000;
			if (ent.mSteps <= 0)
				ent.mSteps = 100;
			SendMessage(ent.mhwndGadget, TBM_SETRANGE, TRUE, MAKELONG(0, ent.mSteps));

			float frange = ent.mMaxValue.v[0] - ent.mMinValue.v[0];
			if (fabsf(frange) > 1e-5f) {
				float fpos = (ent.mValue.v[0] - ent.mMinValue.v[0]) * (float)ent.mSteps / frange;
				float fmax = (float)ent.mSteps;
				if (!(fpos < fmax))
					fpos = fmax;
				if (!(fpos >= 0.0f))
					fpos = 0.0f;

				SendMessage(ent.mhwndGadget, TBM_SETPOS, TRUE, VDRoundToInt(fpos));
			}
		}

		y += mParamUIHeight + mParamUISpacerHeight;
		id += kControlIdWidth;
	}

	// translate UI gadgets
	static const UINT kTranslationIds[]={
		IDC_PREVIEW,
		IDOK,
		IDCANCEL
	};

	const int n = sizeof(kTranslationIds)/sizeof(kTranslationIds[0]);
	int deltaY = y - mPreviewButtonY;

	HDWP hdwp = BeginDeferWindowPos(n);
	RECT r;
	for(int i=0; i<n; ++i) {
		HWND hwndControl = GetDlgItem(mhdlg, kTranslationIds[i]);
		if (!hwndControl)
			continue;

		GetWindowRect(hwndControl, &r);
		MapWindowPoints(NULL, mhdlg, (LPPOINT)&r, 2);
		HDWP hdwpNew = DeferWindowPos(hdwp, hwndControl, NULL, r.left, r.top + deltaY, 0, 0, SWP_NOSIZE|SWP_NOZORDER|SWP_NOACTIVATE);
		if (!hdwpNew)
			break;
		hdwp = hdwpNew;
	}

	EndDeferWindowPos(hdwp);

	if (GetWindowRect(mhdlg, &r) && !(GetWindowLong(mhdlg, GWL_STYLE) & WS_CHILD)) {
		SetWindowPos(mhdlg, NULL, 0, 0, r.right-r.left, r.bottom - r.top + deltaY, SWP_NOMOVE|SWP_NOZORDER|SWP_NOACTIVATE);
	}

	OnSize();
}

void VDShaderFilterDialog::UpdateEnables() {
	EnableWindow(GetDlgItem(mhdlg, IDC_WIDTH), !mConfig.mbKeepSize);
	EnableWindow(GetDlgItem(mhdlg, IDC_HEIGHT), !mConfig.mbKeepSize);
}

void VDShaderFilterDialog::UpdateParameters() {
	if (!mpFilter)
		return;

	int n = mParameterUI.size();
	for(int i=0; i<n; ++i) {
		const ParameterUIEntry& ent = mParameterUI[i];

		mpFilter->SetEffectParameterLive(ent.mName.c_str(), ent.mValue);
	}

	if (mifp)
		mifp->RedoFrame();
}

bool VDShaderFilterShowConfigDialog(VDXHWND hwndParent, VDShaderFilterConfig& cfg, IVDXFilterPreview *ifp, IVDShaderFilter *isf) {
	VDShaderFilterDialog dlg(cfg, ifp, isf);

	return dlg.Show((HWND)hwndParent);
}

IVDShaderFilterConfigDialog *VDShaderFilterCreateConfigDialog(VDXHWND hwndParent, VDShaderFilterConfig& cfg, IVDXFilterPreview *ifp, IVDShaderFilter *isf) {
	VDShaderFilterDialog *dlg = new VDShaderFilterDialog(cfg, ifp, isf);

	if (dlg->ShowModeless((HWND)hwndParent))
		return dlg;

	delete dlg;
	return NULL;
}
