#ifndef f_PREPROCESSOR_H
#define f_PREPROCESSOR_H

#ifdef _MSC_VER
	#pragma once
#endif

#include <map>
#include <vd2/system/vdstl.h>
#include <vd2/system/VDString.h>

#include "lexer.h"

class VDShaderPreprocessor : public IVDShaderErrorOutput {
public:
	VDShaderPreprocessor();
	~VDShaderPreprocessor();

	void Init(const char *src, size_t len, const char *context, IVDShaderLogOutput *pOutput);
	bool GetErrorInfo(VDShaderErrorInfo& errorInfo) { return mLexer.GetErrorInfo(errorInfo); }

	IVDShaderLogOutput *GetLogOutput() const { return mLexer.GetLogOutput(); }
	bool IsErrorFlagSet() const { return mLexer.IsErrorFlagSet(); }

	int Token();
	void Push();

	bool SetError(const char *s);
	bool SetErrorF(const char *format, ...);
	bool SetErrorVF(const char *format, va_list val);
	void ClearErrors();

public:
	const char *mpToken;
	int mTokenLength;
	union {
		int		mIntVal;
		float	mFloatVal;
	};

protected:
	bool ParseDefine();
	int InternalToken();

	bool	mbStartOfLine;
	bool	mbTokenQueued;
	int		mLastToken;

	VDShaderLexer mLexer;

	struct StoredToken {
		int mToken;
		int mTokenOffset;
		int mTokenLength;
		union {
			int mIntVal;
			float mFloatVal;
		};
	};

	struct MacroDefinition {
		typedef vdfastvector<StoredToken> Tokens;
		Tokens mTokens;
		vdfastvector<char>	mTokenHeap;
	};

	typedef std::map<VDStringA, MacroDefinition> Macros;
	Macros mMacros;

	struct ExpandedToken {
		int mToken;
		const char *mpToken;
		int mTokenLength;

		union {
			int		mIntVal;
			float	mFloatVal;
		};
	};

	typedef vdfastvector<ExpandedToken> MacroExpansion;
	MacroExpansion mMacroExpansion;
	int mMacroExpansionIdx;
};

#endif
