// Name: Multi-frame color blend
// Description: Blends the luminance of three adjacent frames together into a color spread.
//
// This filter uses the vd_srcframeoffset annotation on textures in order to
// fetch multiple adjacent frames from the same input stream. A three frame
// window is sampled and each frame converted to monochrome. The mono images
// are then combined to produce colored output.
//

texture srctex0 < float vd_srcframeoffset=-1; >;
texture srctex1 < float vd_srcframeoffset=0; >;
texture srctex2 < float vd_srcframeoffset=1; >;

sampler src0 = sampler_state { texture = (srctex0); };
sampler src1 = sampler_state { texture = (srctex1); };
sampler src2 = sampler_state { texture = (srctex2); };

float4 PS(float2 uv: TEXCOORD0) : COLOR0 {
	float3 px0 = tex2D(src0, uv).rgb;
	float3 px1 = tex2D(src1, uv).rgb;
	float3 px2 = tex2D(src2, uv).rgb;
	
	float3 chromaAxis = { 0.30, 0.59, 0.11 };
	float y0 = dot(px0, chromaAxis);
	float y1 = dot(px1, chromaAxis);
	float y2 = dot(px2, chromaAxis);
	
	float4 result = 0;
	
	result.r = (y0+y1)/2;
	result.g = (y0+y1+y2)/3;
	result.b = (y1+y2)/2;
	
	return result;
}

technique {
	pass {
		PixelShader = compile ps_2_0 PS();
	}
}
