// Ripple distortion
//
// Name: Ripple distortion
// Author: Avery Lee
// Description: Uses a precomputed displacement texture to distort the incoming image.

float4 ComputeDisplacementTexture(float2 uv: TEXCOORD0) : COLOR0 {
	float2 vec = uv - 0.5;
	float x = length(vec)*20;
	float depth = sin(x*x);

	return float4(vec * depth * 0.5 + 0.5, 0, 0);
}

texture disptex <
	string function = "ComputeDisplacementTexture";
	float width = 256;
	float height = 256;
>;

sampler disp = sampler_state {
	texture = (disptex);
	minfilter = linear;
	magfilter = linear;
};

texture vd_srctexture;

sampler src = sampler_state {
	texture = (vd_srctexture);
};

float4 PS(float2 uv: TEXCOORD0) : COLOR0 {
	float2 uv2 = uv + (tex2D(disp, uv).rg - 0.5) * 0.07;

	return tex2D(src, uv2);
}

technique {
	pass {
		PixelShader = compile ps_3_0 PS();
	}
}
