// Name: Texture addressing modes
// Description: Shows the difference between wrap and clamp modes.
//
// The source image is shrunk and rotated, and then tiled across the
// top and bottom halves. The top half uses clamp addressing, while
// the bottom half uses wrapping. With clamp addressing, any texture
// (UV) coordinates are clamped to 0-1, which causes the borders of
// the texture to be extended infinitely and only one image to appear.
// With wrapping, the texture image is tiled in all directions.
//

texture vd_srctexture;

sampler srcClamp = sampler_state {
	Texture = (vd_srctexture);
	AddressU = clamp;
	AddressV = clamp;
};

sampler srcWrap = sampler_state {
	texture = (vd_srctexture);
	AddressU = wrap;
	AddressV = wrap;
};

float4 PS(float2 uv: TEXCOORD0) : COLOR0 {
	float f = 0.707107;
	float2x2 rot = { f, -f, f, f };

	float2 uv2 = mul(uv - float2(0.5, 0.25), rot) * 4 + 0.5;
	float3 pxC = tex2D(srcClamp, uv2).rgb;
	float3 pxW = tex2D(srcWrap, uv2).rgb;
	
	float3 px = uv.y < 0.5 ? pxC : pxW;

	return float4(px, 0);
}

technique {
	pass {
		PixelShader = compile ps_2_0 PS();
	}
}
