#ifndef SCALAR_H
#define SCALAR_H

#ifdef _MSC_VER
	#pragma once
#endif

class IVDShaderLogOutput;

enum {
	kVDShaderInterpOp_end,
	kVDShaderInterpOp_mov,
	kVDShaderInterpOp_add,
	kVDShaderInterpOp_sub,
	kVDShaderInterpOp_mul,
	kVDShaderInterpOp_rsq,
	kVDShaderInterpOp_rcp,
	kVDShaderInterpOp_frc,
	kVDShaderInterpOp_neg,
	kVDShaderInterpOp_min,
	kVDShaderInterpOp_max,
	kVDShaderInterpOp_abs,
	kVDShaderInterpOp_log,
	kVDShaderInterpOp_exp,
	kVDShaderInterpOp_cmp,
	kVDShaderInterpOp_texld,
	kVDShaderInterpOp_unpackb,
	kVDShaderInterpOp_unpackg,
	kVDShaderInterpOp_unpackr,
	kVDShaderInterpOp_unpacka,
	kVDShaderInterpOp_beginloop
};

struct VDShaderInterpConstant {
	int mConstantIndex;
	ptrdiff_t mOffset;
};

struct VDShaderInterpreterInterpolator {
	float	u[4];
	float	dudx[4];
};

struct VDShaderInterpreterSampler {
	const uint32 *mpSrc;
	ptrdiff_t	mPitch;
	float	mFWidth;
	float	mFHeight;
	float	mFWidth0;
	float	mFHeight0;
	uint32	mULimit;
	uint32	mVLimit;
	bool	mbClampU;
	bool	mbClampV;
	bool	mbBilinear;

	void	Init(const uint32 *src, ptrdiff_t pitch, uint32 w, uint32 h, bool clampU, bool clampV, bool bilinear);
	uint32	Sample(float u, float v) const;
};

struct VDShaderInterpreterTransformedInsn {
	int op;
	int dst;
	int src[3];
};

struct VDShaderScalarForm {
	VDShaderScalarForm();
	~VDShaderScalarForm();

	void Clear();

	typedef VDShaderInterpreterTransformedInsn TransformedInsn;
	typedef vdfastvector<TransformedInsn> Insns;
	Insns mInsns;

	vdfastvector<float> mConstants;
	int mTempCount;

	int mInputCount;		// not valid until end

	typedef vdfastvector<int> ConstantBindings;
	ConstantBindings mConstantBindings;

	struct ConstantUploadEntry {
		int mDstIdx;
		ptrdiff_t mSrcOffset;
	};

	typedef vdfastvector<ConstantUploadEntry> ConstantUploads;
	ConstantUploads mConstantUploads;
};

class VDShaderScalarTransformer {
public:
	VDShaderScalarTransformer(VDShaderScalarForm& dst);
	~VDShaderScalarTransformer();

	bool LoadShader(const VDShaderFunction& func, const VDShaderBindingElement *pElements, int numElements, const VDShaderInterpreterSampler *const *pSamplers, int numSamplers, const VDShaderInterpConstant *pConstants, int numConstants, const VDShaderInterpreterInterpolator *interps, uint32 interpCount, float interpStepFactor, IVDShaderLogOutput *out, bool showDisassembly);

protected:
	void mov(int dst, int src);
	void add(int dst, int src1, int src2);
	void sub(int dst, int src1, int src2);
	void mul(int dst, int src1, int src2);
	void min(int dst, int src1, int src2);
	void max(int dst, int src1, int src2);
	void rsq(int dst, int src);
	void rcp(int dst, int src);
	void frc(int dst, int src);
	void log(int dst, int src);
	void exp(int dst, int src);
	void neg(int dst, int src);
	void abs(int dst, int src);
	void cmp(int dst, int src1, int src2, int src3);
	void texld(int dst, int srcu, int srcv, int srcsamp);
	void unpackr(int dst, int src);
	void unpackg(int dst, int src);
	void unpackb(int dst, int src);
	void unpacka(int dst, int src);
	void beginloop();

	void push_insn(int op, int dst, int src1, int src2, int src3 = 0);

	VDShaderScalarForm&	mShader;
};

#endif
