//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "vdfilter.h"

///////////////////////////////////////////////////////////////////////////

VDFilterBase::VDFilterBase(VDXFilterActivation *fa_, const VDXFilterFunctions *ff_)
	: fa(fa_)
	, ff(ff_)
{
}

VDFilterBase::~VDFilterBase() {
}

///////////////////////////////////////////////////////////////////////////

bool VDFilterBase::Init() {
	return true;
}

void VDFilterBase::Start() {
}

void VDFilterBase::Stop() {
}

bool VDFilterBase::Configure(VDXHWND hwnd) {
	return false;
}

void VDFilterBase::GetSettingString(char *buf, int maxlen) {
}

void VDFilterBase::GetScriptString(char *buf, int maxlen) {
}

int VDFilterBase::Serialize(char *buf, int maxbuf) {
	return 0;
}

int VDFilterBase::Deserialize(const char *buf, int maxbuf) {
	return 0;
}

///////////////////////////////////////////////////////////////////////////

void __cdecl VDFilterBase::FilterDeinit   (VDXFilterActivation *fa, const VDXFilterFunctions *ff) {
	reinterpret_cast<VDFilterBase *>(fa->filter_data)->~VDFilterBase();
}

int  __cdecl VDFilterBase::FilterRun      (const VDXFilterActivation *fa, const VDXFilterFunctions *ff) {
	VDFilterBase *pThis = reinterpret_cast<VDFilterBase *>(fa->filter_data);

	pThis->fa		= const_cast<VDXFilterActivation *>(fa);

	return !pThis->Run();
}

long __cdecl VDFilterBase::FilterParam    (VDXFilterActivation *fa, const VDXFilterFunctions *ff) {
	VDFilterBase *pThis = reinterpret_cast<VDFilterBase *>(fa->filter_data);

	pThis->fa		= fa;

	return pThis->GetParams();
}

int  __cdecl VDFilterBase::FilterConfig   (VDXFilterActivation *fa, const VDXFilterFunctions *ff, VDXHWND hwnd) {
	VDFilterBase *pThis = reinterpret_cast<VDFilterBase *>(fa->filter_data);

	pThis->fa		= fa;

	return !pThis->Configure(hwnd);
}

int  __cdecl VDFilterBase::FilterStart    (VDXFilterActivation *fa, const VDXFilterFunctions *ff) {
	VDFilterBase *pThis = reinterpret_cast<VDFilterBase *>(fa->filter_data);

	pThis->fa		= fa;

	pThis->Start();
	return 0;
}

int  __cdecl VDFilterBase::FilterEnd      (VDXFilterActivation *fa, const VDXFilterFunctions *ff) {
	VDFilterBase *pThis = reinterpret_cast<VDFilterBase *>(fa->filter_data);

	pThis->fa		= fa;

	pThis->Stop();
	return 0;
}

bool __cdecl VDFilterBase::FilterScriptStr(VDXFilterActivation *fa, const VDXFilterFunctions *ff, char *buf, int buflen) {
	VDFilterBase *pThis = reinterpret_cast<VDFilterBase *>(fa->filter_data);

	pThis->fa		= fa;

	pThis->GetScriptString(buf, buflen);

	return true;
}

void __cdecl VDFilterBase::FilterString2  (const VDXFilterActivation *fa, const VDXFilterFunctions *ff, char *buf, int maxlen) {
	VDFilterBase *pThis = reinterpret_cast<VDFilterBase *>(fa->filter_data);

	pThis->fa		= const_cast<VDXFilterActivation *>(fa);

	pThis->GetSettingString(buf, maxlen);
}

int  __cdecl VDFilterBase::FilterSerialize    (VDXFilterActivation *fa, const VDXFilterFunctions *ff, char *buf, int maxbuf) {
	VDFilterBase *pThis = reinterpret_cast<VDFilterBase *>(fa->filter_data);

	pThis->fa		= fa;

	return pThis->Serialize(buf, maxbuf);
}

void __cdecl VDFilterBase::FilterDeserialize  (VDXFilterActivation *fa, const VDXFilterFunctions *ff, const char *buf, int maxbuf) {
	VDFilterBase *pThis = reinterpret_cast<VDFilterBase *>(fa->filter_data);

	pThis->fa		= fa;

	pThis->Deserialize(buf, maxbuf);
}
